/* PolyView.java
 *
 * Jim McBeath, May 24, 2002
 * Merged from ListView and SingleView of June 7, 2002
 */

package jimmc.roots;

import jimmc.swing.ButtonAction;
import jimmc.swing.CheckBoxAction;
import jimmc.swing.ComboBoxAction;
import jimmc.swing.ComponentPrintDialog;
import jimmc.swing.GridBagger;
import jimmc.swing.MenuAction;
import jimmc.swing.JsTextField;
import jimmc.treepar.Layout;
import jimmc.treepar.LayoutPanel;
import jimmc.util.ArrayUtil;
import jimmc.util.MoreException;

import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.Box;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.GridBagConstraints;
import java.awt.print.PrinterException;
import java.util.Vector;

/** PolyView provides a view into either a list or a single record.
 */
public class PolyView extends RootsFrame {
	/** Our database. */
	protected DatabaseModel databaseModel;

	/** Our model. */
	protected PolyModel model;

	/** The box with info about a list of records. */
	protected Component listInfoArea;

	/** The box with info about one record. */
	protected Component recordInfoArea;

	/** Where we display the count of items in our list when > 1. */
	protected JLabel countField;

	/** Where we display the key value for our record when one record. */
	protected JLabel keyField;

	/** Where we display the type of our record when one record. */
	protected JLabel typeField;

	/** Where we display the one-line summary of our record when one. */
	protected JLabel summaryField;

	/** What to do with the selections from the selection buttons. */
	protected ComboBoxAction selectionOpField;

	/** Contains either the text panel or the layout panel. */
	protected TextOrLayoutPanel outputPanel;

	/** The output type currently selected. */
	protected JComboBox outputChoiceField;

	/** The include-keys toggle. */
	protected JCheckBox includeKeysField;

	/** The number of Poly windows we have created.
	 * This is used to get a unique ID for each one. */
	static int windowCount;

	/** The ID number for this Poly window. */
	protected int windowId;

	/** Create a new view for many or one record. */
	public PolyView(DatabaseModel db, String[] keys) {
		super("Poly");
		windowId = ++windowCount;
		this.databaseModel = db;
		setJMenuBar(createMenuBar());
		initForm();
		addWindowListener();
		setModel(db.getPolyModel(keys));
		pack();
	}

	/** Set the model we are viewing. */
	public void setModel(PolyModel m) {
		model = m;
		String title = "roots list #"+windowId;
		int keyCount = m.getCount();
		if (keyCount==1) {
			String[] keys = m.getKeys();
			keyField.setText(keys[0]);
			typeField.setText(m.getType());
			summaryField.setText(m.getSummary());
			title = title+": "+keys[0];
		} else {
			countField.setText(Integer.toString(m.getCount()));
		}
		setTitle(title);
		recordInfoArea.setVisible(keyCount==1);
		listInfoArea.setVisible(keyCount!=1);
		setOutputDisplay();
		validate();
	}

	/** Set the output field. */
	protected void setOutputDisplay() {
		if (includeKeysField==null)
			return;	/* this happens during init */
		boolean includeKeys = includeKeysField.isSelected();
		String outputFormat =
			(String)outputChoiceField.getSelectedItem();
		if (outputFormat.startsWith("G")) {
			Layout layout = model.getOutputLayout(
					outputFormat,includeKeys);
			outputPanel.setLayout(layout);
		} else {
			String text = model.getOutputText(
					outputFormat,includeKeys);
			outputPanel.setText(text);
		}
	}

	/** Create our menu bar. */
	protected JMenuBar createMenuBar() {
		JMenuBar mb = new JMenuBar();
		mb.add(createFileMenu());
		//JMenu helpMenu = createHelpMenu();
		//mb.add(helpMenu);
		//mb.setHelpMenu(helpMenu);
		return mb;
		/* TBD - add a select menu with things like ancestors to
		 * N generations, descendants to N generations, whatever
		 * else seems reasonable.
		 */
		/* TBD - add a menu command to save the large text area
		 * to a file. */
	}

	/** Create our File menu. */
	protected JMenu createFileMenu() {
		JMenu m = new JMenu("File");
		MenuAction mi;
		mi = new MenuAction("Save Output As...") {
			public void action() {
				processSaveOutputAs();
			}
		};
		m.add(mi);
		mi = new MenuAction("Show Display Command") {
			public void action() {
				processShowDisplayCommand();
			}
		};
		mi = new MenuAction("Print...") {
			public void action() {
				processPrint();
			}
		};
		m.add(mi);
		mi = new MenuAction("Close") {
			public void action() {
				processClose();
			}
		};
		m.add(mi);
		return m;
	}

	/** Create the body of our form. */
	protected void initForm() {
		GridBagger gb = new GridBagger(getContentPane());
		gb.gbc.fill = GridBagConstraints.HORIZONTAL;
		gb.gbc.anchor = GridBagConstraints.WEST;
		gb.add(listInfoArea=createListInfoArea());
		gb.nextRow();
		gb.add(recordInfoArea=createRecordInfoArea());
		gb.nextRow();
		gb.add(createSelectionArea());
		gb.nextRow();
		gb.gbc.weightx = 1;
		gb.gbc.weighty = 1;
		gb.gbc.fill = GridBagConstraints.BOTH;
		gb.add(createOutputArea());
	}
	
	/** Create the record info box at the top of the form. */
	protected Component createRecordInfoArea() {
		JPanel p = new JPanel();
		p.setBorder(new TitledBorder(new EtchedBorder(),"Record"));
		GridBagger gb = new GridBagger(p);
		gb.gbc.fill = GridBagConstraints.HORIZONTAL;
		gb.gbc.anchor = GridBagConstraints.WEST;
		gb.add(new JLabel("Key:"));
		gb.add(keyField=new JLabel());
		gb.nextRow();
		gb.add(new JLabel("Type:"));
		gb.add(typeField=new JLabel());
		gb.nextRow();
		gb.add(new JLabel("Summary:"));
		gb.gbc.weightx = 1;
		gb.add(summaryField=new JLabel());
		return p;
	}

	/** Create the list info box at the top of the form. */
	protected Component createListInfoArea() {
		JPanel p = new JPanel();
		p.setBorder(new TitledBorder(new EtchedBorder(),"List"));
		GridBagger gb = new GridBagger(p);
		gb.gbc.anchor = GridBagConstraints.WEST;
		gb.add(new JLabel("Count:"));
		gb.gbc.fill = GridBagConstraints.HORIZONTAL;
		gb.gbc.weightx = 1;
		gb.add(countField=new JLabel());
		gb.gbc.weightx = 0;
		//Add two filler lines to make our info box
		//the same size as the Record info box.
		//This prevents stuff from jumping around when we
		//switch between Record and List info boxes.
		gb.nextRow();
		gb.add(new JLabel(" "));
		gb.nextRow();
		gb.add(new JLabel(" "));
		return p;
	}

	/** Create the selection buttons. */
	protected Component createSelectionArea() {
		JPanel p = new JPanel();
		p.setBorder(new TitledBorder(new EtchedBorder(),"Select"));
		GridBagger gb = new GridBagger(p);
		gb.gbc.gridwidth = 4;
		selectionOpField = new ComboBoxAction();
		//See processSelect for the case statement that
		//corresponds to these messages
		selectionOpField.addItem("Replace Current Selection");
		selectionOpField.addItem("Add To Current Selection");
		selectionOpField.addItem("Open Selection In New Window");
		//selectionOpField.addItem("Replace With A Single Selection");
		Box selOpBox = Box.createHorizontalBox();
		selOpBox.add(new JLabel("Action: "));
		selOpBox.add(selectionOpField);
		gb.add(selOpBox);
		gb.nextRow();
		gb.gbc.gridwidth = 1;
		gb.gbc.fill = GridBagConstraints.HORIZONTAL;
		gb.add(new SelectButton("Father"));
		gb.add(new SelectButton("Mother"));
		gb.add(new SelectButton("Parents"));
		gb.add(new SelectButton("Ancestors","recurse parents"));
		gb.gbc.weightx = 1;
		gb.add(new JLabel(""));	/* filler column */
		gb.gbc.weightx = 0;
		gb.nextRow();
		gb.add(new SelectButton("Husband"));
		gb.add(new SelectButton("Wife"));
		gb.add(new SelectButton("Children"));
		gb.add(new SelectButton("Descendents","recurse children"));
		gb.nextRow();
		gb.add(new SelectButton("Marriage"));
		gb.add(new SelectButton("ParentUnion"));
		gb.add(new SelectButton("Current",""));

		/* Put in a general purpose select text box to allow
		 * easy selection of things like "plus recurse4 father"
		 */
		gb.nextRow();
		gb.gbc.gridwidth = 4;
		Box b = Box.createHorizontalBox();
		b.add(new JLabel("Direct select: "));
		b.add(new JsTextField(20) {
			public void action() {
				processSelect(getText());
			}
		});
		gb.add(b);

		return p;
	}

	/** Create the output area. */
	protected Component createOutputArea() {
		JPanel p = new JPanel();
		p.setBorder(new TitledBorder(new EtchedBorder(),"Output"));
		GridBagger gb = new GridBagger(p);
		gb.add(new JLabel("Display:"));
		outputChoiceField = new ComboBoxAction() {
			public void action() {
				setOutputDisplay();
			}
		};
		outputChoiceField.addItem("list");
		outputChoiceField.addItem("atree");
		outputChoiceField.addItem("dtree");
		outputChoiceField.addItem("fampage");
		outputChoiceField.addItem("gedcom");
		outputChoiceField.addItem("ncv");
		outputChoiceField.addItem("Tctree");
		outputChoiceField.addItem("Gctree");
		outputChoiceField.setSelectedItem("list");
		gb.add(outputChoiceField);
		includeKeysField = new CheckBoxAction("Include Keys") {
			public void action() {
				setOutputDisplay();
			}
		};
		includeKeysField.setSelected(true);
		gb.add(includeKeysField);
		gb.nextRow();
		gb.gbc.gridwidth = 4;
		gb.gbc.weightx = 1;
		gb.gbc.weighty = 1;
		gb.gbc.fill = GridBagConstraints.BOTH;
		outputPanel = new TextOrLayoutPanel() {
			public void processTextSelect(String key) {
				PolyView.this.processSelect(key);
			}
			public void processTextRemove(String key) {
				PolyView.this.processRemove(key);
			}
			public Dimension getPreferredSize() {
				return new Dimension(500,300);
			}
		};
		gb.add(outputPanel);

		return p;
	}

	/** Change the selection. */
	public void processSelect(String sel) {
		String[] newSel = model.getSelection(sel);
		if (newSel==null || newSel.length==0) {
			errorDialog("No match");
				/* nothing selected, don't switch */
			return;
		}
		int selType = selectionOpField.getSelectedIndex();
		switch (selType) {
		//See createSelectionArea for the messages that
		//correspond to these cases.
		case 1:		//add to current
			newSel = ArrayUtil.union(model.getKeys(),newSel);
			//FALL THROUGH, replace with the combined set
		case 0:		//replace current
			setModel(databaseModel.getPolyModel(newSel));
			break;
		case 2:		//new window
			(new PolyView(databaseModel,newSel)).show();
			break;
		case 3:		//replace with a single selection
			/* multiple items, put up a choice list
			 * and let him pick one or cancel,
			 * or open up a ListView with the items.
			 */
			String oneSel = processListSelection(newSel);
			/* If he picked one of them, switch to that record;
			 * else ignore it. */
			if (oneSel!=null)
				setModel(databaseModel.getPolyModel(oneSel));
			break;
		}
	}

	/** Remove an item from the current selection. */
	public void processRemove(String key) {
		String[] keys = model.getKeys();
		if (keys==null || keys.length==0)
			return;		//nothing from which to remove
		if (key==null)
			return;		//no key to remove
		int k;
		for  (k=0; k<keys.length; k++) {
			if (key.equals(keys[k]))
				break;	//found it
		}
		if (k>=keys.length)
			return;		//not found
		String[] newKeys = new String[keys.length - 1];
		if (k>0)
			System.arraycopy(keys,0,newKeys,0,k);
		if (k<keys.length-1)
			System.arraycopy(keys,k+1,newKeys,k,keys.length-k-1);
		//Replace our selection with this new selection
		setModel(databaseModel.getPolyModel(newKeys));
	}

	/** Put up a choicelist of items, let the user select one.
	 * @return the selected key, or null if no key was selected.
	 *         This happens if the user Cancels, or selects the option
	 *         to open up a ListView.
	 */
	protected String processListSelection(String[] keys) {
		String prompt = "Select which record to switch to.\n"+
			"Number of records: "+keys.length;
		String[] keySummaries = databaseModel.getKeySummaries(keys);
		JList list = new JList(keySummaries);
		JScrollPane pane = new JScrollPane(list);
		Object[] msgs = { prompt, pane };
		Object[] options = { "OK", "Cancel", "ListView" };
		int n = JOptionPane.showOptionDialog(this,msgs,"Select",
			JOptionPane.DEFAULT_OPTION,
			JOptionPane.QUESTION_MESSAGE,null,options,options[0]);
		if (n==0)
			return keys[list.getSelectedIndex()];
		if (n==2)
			(new PolyView(databaseModel,keys)).show();
		return null;
	}

	/** Process the File->ShowDisplayCommand command. */
	protected void processShowDisplayCommand() {
		boolean includeKeys = includeKeysField.isSelected();
		String outputFormat =
			(String)outputChoiceField.getSelectedItem();
		String text = model.getDisplayCommand(outputFormat,includeKeys);
		infoDialog(text);
	}

	/** Process the File->SaveOutputAs command. */
	protected void processSaveOutputAs() {
		/* TBD - check if using layout instead of text. */
		String text = outputPanel.getText();
		if (text.length()==0) {
			errorDialog("No text to save.");
			return;
		}
		String filename = fileDialog("Save to file:");
		if (filename==null)
			return;
		saveTextToFile(text,filename);
	}

	/** Process the File->Print command. */
	protected void processPrint() {
		Component c = outputPanel.getPrinterComponent();
		ComponentPrintDialog d = new ComponentPrintDialog(this,
			databaseModel.getBase(),c);
		if (!d.printDialog())
			return;		//cancelled
		try {
			d.print();
		} catch (PrinterException ex) {
			throw new MoreException(ex);
		}
	}

	/** Close the window. */
	protected void processClose() {
		/* TBD - notify the model. */
		super.processClose();
	}

	/* A button to change the selection */
	class SelectButton extends ButtonAction {
		String sel;
		public SelectButton(String label) {
			this(label,label.toLowerCase());
		}
		public SelectButton(String label, String sel) {
			super(label);
			this.sel = sel;
		}
		public void action () {
			PolyView.this.processSelect(sel);
		}
	}
}

/* end */
