/* StringParser.java
 * Copyright 1998 Jim McBeath
 *
 * Jim McBeath, July 7, 1998
 */

package jimmc.roots;

import java.util.Vector;

/** String parser parses a string into a series of smaller strings using
 * shell-style rules: unquoted strings are broken at spaces, or strings can
 * be quoted using double or single quotes.  Backslash can be used to
 * quote a single character, either within or outside of a string.
 */
public class StringParser {
	/** The string we are parsing. */
	String input;
	/** Our location within the string. */
	int charPos;

	/** Convenience routine to parse a string. */
	public static String[] parse(String input) {
		return new StringParser(input).parse();
	}

	/** Create a string parser. */
	public StringParser(String input) {
		this.input = input;
		charPos = 0;
	}

	/** Parse our string into an array of strings. */
	public String[] parse() {
		Vector wordVector = new Vector();
		String s;
		while ((s=getNextString())!=null) {
			wordVector.addElement(s);
		}
		String[] words = new String[wordVector.size()];
		wordVector.copyInto(words);
		return words;
	}

	/** Get the next string from the input. */
	public String getNextString() {
		int c = nextCharAfterWhiteSpace();
		if (c<0)
			return null;		/* end of string */

		/* build the string */
		StringBuffer sb = new StringBuffer();
		do {
			if (c=='\'' || c=='"')
				addQuotedString(sb,c);
			else if (c=='\\') {
				c = nextChar();
				if (c!=-1)
					sb.append((char)c);
			}
			else
				sb.append((char)c);
			c = nextChar();
		} while (c>=0 && !Character.isWhitespace((char)c));
		return sb.toString();
	}
	
	/** Add characters within quotes to the buffer. */
	protected void addQuotedString(StringBuffer sb, int quoteChar) {
		int c = nextChar();
		do {
			if (c==quoteChar)
				return;		/* done with quoted string */
			else
				sb.append((char)c);
			c = nextChar();
		} while (c!=-1);
	}

	/** Skip over white space, return next char.
	 * @return first character after whitespace, or -1 on EOS.
	 */
	protected int nextCharAfterWhiteSpace() {
		while (true) {
			int c = nextChar();
				/* get next char, -1 on EOS */
			if (!Character.isWhitespace((char)c))
				return c;
		}
	}

	/** Get the next character from the string. */
	protected int nextChar() {
		if (charPos>=input.length())
			return -1;	/* EOS */
		int c = input.charAt(charPos++);
		//System.out.println("nextChar "+(char)c);
		return c;
	}

	/** Return true if we are at the end of the string. */
	protected boolean isEos() {
		return (charPos>=input.length());
	}

	/** A main routine for testing. */
	public static void main(String[] args) {
		String[] parsed = StringParser.parse(args[0]);
		System.out.println("length="+parsed.length);
		for (int i=0; i<parsed.length; i++) {
			System.out.println(parsed[i]);
		}
	}
}

/* end */
