/* Layout.java
 *
 * Jim McBeath, July 8, 1998
 */

package jimmc.treepar;

import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.io.PrintWriter;
import java.util.Dictionary;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/** Layout is the root object containing all of the data used to do a
 * place and route with the treepar package.
 * <p>
 * There are three phases of use for Layout, which must be done in order:
 * <ul>
 * <li>Setup - During the setup phase, the application must create a set of 
 *     <a href="jimmc.treepar.Box.html">Box</a> and
 *     <a href="jimmc.treepar.Connector.html">Connector</a>
 *     objects in the Layout.
 * <li>Place-and-Route - Once the setup phase is complete, a call to
 *     <a href="#placeAndRoute">placeAndRoute</a>
 *     will perform the place and route.
 *     This method must only be called once after setup.
 * <li>Paint - the
 *     <a href="#paint">paint</a> method will draw the layout to the specified
 *     Graphics object.  Paint can be called any number of times.
 * </ul>
 * <p>
 * Layout uses a channel-based algorithm.  The layout is optimized for
 * looks (as opposed to density, as is typical in a CAD application),
 * and is intended for simple tree-structured data, as is commonly found
 * in a family tree.  It will handle arbitrary network-stuctured data,
 * but may not make a very nice looking plot.
 * It will insert feedthroughs when necessary, but is not particularly
 * intelligent about where in a row a feedthrough should go.
 * <p>
 * The algorithm uses the following additional object types:
 * <ul>
 * <li><a href="jimmc.treepar.Row.html">Row</a> - a row of Box objects,
 *     or a channel row with routing tracks.
 * <li><a href="jimmc.treepar.Box.html">Box</a> - a fixed-size rectangular
 *     area with a set of Connector objects, within which is descriptive text
 *     (typically information about a family or individual).
 * <li><a href="jimmc.treepar.Connector.html">Connector</a> - a point on a Box
 *     that connects to a Net.  Lines are drawn from the connector to other
 *     connectors on the same net.
 * <li><a href="jimmc.treepar.Net.html">Net</a> - a collection of connectors
 *     with the same name.
 * </ul>
 * <p>
 * Layout can work with either horizontal rows or vertical rows.
 * In many places, the algorithm is simplified by being
 * able to ignore this selection.  In order to reduce confusion, terms such
 * as width, height, x, and y are avoided.  Instead, a different set of terms
 * are used.
 * <p>
 * For the dimensions of a row, thickness and length are used instead of
 * width and height.
 * The long dimension of the row (in the typical case) is
 * called the length, and the short dimension is called the thickness.
 * Thus if the rows are oriented horizontally,
 * the thickness of a row is its height, and the
 * length of the row is its width; whereas if the rows are oriented vertically,
 * then the thickness is the width, and the length is the height.  By using
 * thickness and length rather than width and height the algorithm can in most
 * places ignore the actual orientation of the rows.
 * <p>
 * For point coordinates, r and s are used instead of x any y.
 * The r coordinate
 * is the position of the point along the length of the rows, whereas the s
 * coordinate is the position of the point along the thickness of the rows.
 * Thus if the rows are oriented horizontally, then r=x and s=y; whereas if
 * the rows are oriented vertically, then r=y and s=x.
 * <p>
 * For directions, the terms Rmore, Rless, Smore, and Sless are used,
 * referring to the directions of greater or lesser values of the r and s
 * coordinates.
 * If the rows are oriented horizontally, then Rmore=east, Rless=west,
 * Smore=north, and Sless=south; whereas if the rows are oriented vertically,
 * then Rmore=north, Rless=south, Smore=east, and Sless=west.
 * <p>
 * Each row is assigned a row number.  The channels between rows also have
 * row numbers.  All rows have even row numbers, whereas all channels have
 * odd row numbers.
 * <p>
 * The (x,y) coordinate system has it's origin in the lower left, as with
 * standard mathematical coordinates.
 */
public class Layout {
    /* Constants. */

	/** Constant for connector edges. */
	public static final int NORTH = 1;

	/** Constant for connector edges. */
	public static final int SOUTH = 2;

	/** Constant for connector edges. */
	public static final int EAST = 3;

	/** Constant for connector edges. */
	public static final int WEST = 4;

	/** Constant for routing angle. */
	public static final int LEFT = 5;

	/** Constant for routing angle. */
	public static final int RIGHT = 6;

	/** Constant referring to
	 * the first step in the place-and-route process: collecting the
	 * application-specified data into a form which is easier to process.
	 * <p>
	 * The order of the steps is:
	 * <ul>
	 * <li><a href="#STEP_SETUP_DONE">STEP_SETUP_DONE</a>
	 * <li><a href="#STEP_ROW_NUMBERS">STEP_ROW_NUMBERS</a>
	 * <li><a href="#STEP_FEEDTHROUGHS">STEP_FEEDTHROUGHS</a>
	 * <li><a href="#STEP_ROWS">STEP_ROWS</a>
	 * <li><a href="#STEP_ROW_SPACE">STEP_ROW_SPACE</a>
	 * <li><a href="#STEP_ROW_ORDER">STEP_ROW_ORDER</a>
	 * <li><a href="#STEP_ROW_POSITION">STEP_ROW_POSITION</a>
	 * <li><a href="#STEP_ROUTE">STEP_ROUTE</a>
	 * <li><a href="#STEP_BOUNDS">STEP_BOUNDS</a>
	 * </ul>
	 */
	public static final int STEP_SETUP_DONE = 1;

	/** Constant referring to the step where row numbers are assigned. */
	public static final int STEP_ROW_NUMBERS = 2;

	/** Constant referring to the step where feedthroughs are created. */
	public static final int STEP_FEEDTHROUGHS = 3;

	/** Constant referring to the step where rows are allocated. */
	public static final int STEP_ROWS = 4;

	/** Constant referring to the step where row thicknesses 
	 * are calculated.*/
	public static final int STEP_ROW_SPACE = 5;

	/** Constant referring to the step where boxes within rows 
	 * are ordered.*/
	public static final int STEP_ROW_ORDER = 6;

	/** Constant referring to the step where row positions 
	 * are calculated. */
	public static final int STEP_ROW_POSITION = 7;

	/** Constant referring to the step where routing is done. */
	public static final int STEP_ROUTE = 8;

	/** Constant referring to the step where bounding rectangle 
	 * is calculated. */
	public static final int STEP_BOUNDS = 9;

	/** Constant referring to all steps. */
	public static final int STEP_ALL = 10;

    /* These variables are set during the setup phase. */

	/** The set of Box objects added during the setup phase.
	 * Boxes are automatically added to this list when created
	 * with the Box(Layout) constructor.
	 */
	protected Vector boxVector;

	/** True if rows are vertical; false if horizontal.
	 * Use setVertical to set this value.
	 */
	protected boolean verticalRows;

	/** The root box to use for the place algorithm.  May be null.
	 * This only matters if the network is not a simple tree.
	 */
	protected Box rootBox;

	/** True to use a greedy routing algorithm for track assignment. */
	protected boolean useGreedyRoute = true;

	/** The height of a character (actually, spacing between baselines). */
	protected int lineHeight = 14;

	/** The width of one character. */
	protected int characterWidth = 9;

	/** The spacing between rows (S dimension). */
	protected int interRowSpace = 14;

	/** The default spacing between boxes (R dimension) */
	protected int interBoxSpace = 14;

	/** The amount of space between tracks in a routing channel. */
	protected int trackSpace = 14;

    /* These variables are set during the setup phase, but not
     * directly by the user.
     */

	/** NORTH if horizontal rows; EAST if vertical rows (Smore).
	 * Set by setVertical.
	 */
	protected int NorE;

	/** SOUTH if horizontal rows; WEST if vertical rows (Sless).
	 * Set by setVertical.
	 */
	protected int SorW;

    /* These variables are set during the route phase. */

	/** The set of Box objects used during the layout process. */
	protected Box[] boxes;

	/** The set of Net objects, while collecting nets. */
	protected Dictionary netsDict;

	/** The set of Net objects used during the layout process. */
	protected Net[] nets;

	/** The set of Connector objects used during the layout process. */
	protected Connector[] connectors;

	/** The set of Row objects in the layout. */
	protected Row[] rows;

	/** The maximum length of a row. */
	protected int maxRowLength;

	/** Our bounds, in which all drawable items lie. */
	protected BoundsRectangle bounds;

    /* methods. */

	/** Create an empty Layout. */
	public Layout(boolean verticalRows) {
		setRowOrientation(verticalRows);
		boxVector = new Vector();
	}
	
	/** Set the orientation of the rows.
	 * This method may be called at any time during the setup phase.
	 * @param vertical true if the typically long dimension of the rows
	 *        are vertical; false if horizontal.
	 */
	public void setRowOrientation(boolean vertical) {
		this.verticalRows = vertical;
		if (vertical) {
			NorE = EAST;	/* Smore */
			SorW = WEST;	/* Sless */
		} else {
			NorE = NORTH;	/* Smore */
			SorW = SOUTH;	/* Sless */
		}
	}

	/** Set the lineHeight. */
	public void setLineHeight(int lineHeight) {
		this.lineHeight = lineHeight;
	}

	/** Get the lineHeight. */
	public int getLineHeight() {
		return lineHeight;
	}

	/** Set the character width. */
	public void setCharacterWidth(int characterWidth) {
		this.characterWidth = characterWidth;
	}

	/** Get the character width. */
	public int getCharacterWidth() {
		return characterWidth;
	}

	/** Add a box to the list of boxes in the layout.
	 * This is called from the Box constructor.
	 */
	protected void addBox(Box box) {
		boxVector.addElement(box);
		boxes = null;
	}

	/** After all calls to addBox, call this method to indicate
	 * that the setup phase is complete.
	 * This method creates the internal array of boxes used during
	 * calculations.
	 */
	public void setupDone() {
		/* Setup the boxes array. */
		if (boxes==null) {
			boxes = new Box[boxVector.size()];
			boxVector.copyInto(boxes);
		}
		/* Note setup is complete on all boxes. */
		for (int b=0; b<boxes.length; b++)
			boxes[b].setupDone();

		/* Count the connectors on all boxes. */
		int cn = 0;
		for (int b=0; b<boxes.length; b++)
			cn += boxes[b].connectors.length;
		/* Create the single array of all connectors. */
		connectors = new Connector[cn];
		cn = 0;
		for (int b=0; b<boxes.length; b++) {
			System.arraycopy(boxes[b].connectors,0,
					connectors,cn,
					boxes[b].connectors.length);
			cn += boxes[b].connectors.length;
		}

		/* Create the nets, set the net links in the connectors. */
		netsDict = new Hashtable();
		for (int c=0; c<connectors.length; c++)
			connectors[c].linkNet();

		/* Create the single array of all nets. */
		nets = new Net[netsDict.size()];
		Enumeration netsEnum = netsDict.elements();
		int nn = 0;
		while (netsEnum.hasMoreElements())
			nets[nn++] = (Net)netsEnum.nextElement();
		/* Set up the connectors array in all nets. */
		for (int n=0; n<nets.length; n++)
			nets[n].connectorsDone();
	}

	/** Find a net by name.
	 * If not found, creates a new net and adds it to the layout.
	 * @return the net.
	 */
	protected Net findOrCreateNet(String name) {
		Net net = (Net)netsDict.get(name);
		if (net==null) {
			net = new Net(this,name);
			netsDict.put(name,net);
		}
		return net;
	}

	/** Find a net by name.
	 * @return the net, or null if not found.
	 */
	protected Net findNet(String name) {
		Net net = (Net)netsDict.get(name);
		return net;
	}

	/** Do the complete place and route.
	 * @param box the box to be used as the root of the layout process.
	 *        May be null.
	 */
	public void placeAndRoute() {
		placeAndRoute(STEP_ALL);
	}

	/** Do a partial place-and-route, finishing the step specified.
	 * @param box The root box for the route; may be null.
	 * @param step The last
	 *        <a href="#STEP_SETUP_DONE">step</a> to execute.
	 *        Useful for debugging.
	 */
	public void placeAndRoute(int step) {
		setupDone();
		if (step<=STEP_SETUP_DONE)
			return;

		setRowNumbers(rootBox);		/* select into rows */
		if (step<=STEP_ROW_NUMBERS)
			return;

		reFeed();	/* re-do feedthroughs */
		if (step<=STEP_FEEDTHROUGHS)
			return;

		createRows();	/* create rows, put boxes and nets into them. */
		if (step<=STEP_ROWS)
			return;

		rowSpace();	/* calculate thickness and S of each row */
		if (step<=STEP_ROW_SPACE)
			return;

		rowOrder(rootBox); /* calculate the order of boxes w/in rows */
		if (step<=STEP_ROW_ORDER)
			return;

		rowPosition();	/* set the position of each row */
		if (step<=STEP_ROW_POSITION)
			return;

		route();	/* do the channel route */
		if (step<=STEP_ROUTE)
			return;

		setBounds();	/* calculate the bounding rectangle */
	}

	/** Find a box by name.
	 * @param boxName the name of the box.
	 * @return the box, or null if not found.
	 */
	protected Box findBox(String boxName) {
		for (int b=0; b<boxes.length; b++) {
			if (boxes[b].name.equals(boxName))
				return boxes[b];
		}
		return null;
	}

	/** Go through all of the boxes and determine which row each
	 * box goes into.
	 * @param box the root of the row select process.  
	 *        This only matters if the network is not a simple tree.
	 */
	public void setRowNumbers(Box box) {
		/* Clear out the old row numbers in all boxes and nets. */
		for (int b=0; b<boxes.length; b++)
			boxes[b].clearRowNumber();
		for (int n=0; n<nets.length; n++)
			nets[n].clearRowNumber();

		/* If there is a root box, start with it. */
		if (box!=null)
			box.setRowNumber(0);
		/* Make sure all boxes are selected into rows. */
		for (int b=0; b<boxes.length; b++)
			boxes[b].setRowNumber(0);
	}

	/** Redo feedthrougs. */
	public void reFeed() {
		unFeed();		/* undo old feedthroughs */
		feed();			/* make new feedthroughs */
	}

	/** Undo feedthroughs. */
	public void unFeed() {
		/* TBD - need to do this if we run through par twice */
	}

	/** Create feedthroughs. */
	public void feed() {
		/* Split nets which span rows. */
		for (int n=0; n<nets.length; n++)
			nets[n].split();
		/* Clear the connections on all nets. */
		for (int n=0; n<nets.length; n++)
			nets[n].clearConnectors();
		/* Reattach nets to connectors. */
		for (int c=0; c<connectors.length; c++)
			connectors[c].linkNet();
		/* Re-Set up the connectors array in all nets. */
		for (int n=0; n<nets.length; n++)
			nets[n].connectorsDone();
	}

	/** Figure out how many rows we need, create them, and add the
	 * boxes and nets to their rows.
	 */
	public void createRows() {
		/* Find the maximum and minimum row numbers from all
		 * boxes and nets. */
		int maxRownum = 0;
		int minRownum = 0;
		for (int b=0; b<boxes.length; b++) {
			int rn = boxes[b].rowNumber;
			if (b==0 || rn>maxRownum)
				maxRownum = rn;
			if (b==0 || rn<minRownum)
				minRownum = rn;
		}
		for (int n=0; n<nets.length; n++) {
			int rn = nets[n].rowNumber;
			if (rn>maxRownum)
				maxRownum = rn;
			if (rn<minRownum)
				minRownum = rn;
		}

		/* Allocate the new set of rows. */
		int rowCount = maxRownum - minRownum + 1;
		rows = new Row[rowCount];
		for (int rx=0; rx<rows.length; rx++)
			rows[rx] = new Row(this,rx+minRownum);

		/* add all boxes and nets into their rows */
		for (int b=0; b<boxes.length; b++) {
			int rn = boxes[b].rowNumber;
			Row row = rows[rn-minRownum];
			boxes[b].setRow(row);
			row.addBox(boxes[b]);
		}
		for (int n=0; n<nets.length; n++) {
			int rn = nets[n].rowNumber;
			Row row = rows[rn-minRownum];
			nets[n].setRow(row);
			row.addNet(nets[n]);
		}
		/* Note we are done adding stuff to the rows. */
		for (int r=0; r<rows.length; r++)
			rows[r].setupDone();
	}

	/** Calculate the thickness and S coordinate of each row.
	 * Set the S coordinat of each box in the row, then
	 * recalculate the positions of the connectors.
	 */
	public void rowSpace() {
		/* calculate the thickness and position (S coord) of each row */
		int pos = 0;
		for (int r=0; r<rows.length; r++) {
			Row row = rows[r];
			row.setS(pos);
			pos += interRowSpace;
			pos += row.calculateThickness();
		}

		/* Set the S coordinate for each box to its row's S coordinate.
		 * The R coordinate gets set later.
		 */
		for (int b=0; b<boxes.length; b++)
			boxes[b].setS();
	}

	/** Calculate the ordering of the boxes within the rows.
	 * A depth-first network traversal is done, and a box is assigned
	 * an ordinal position with the its row the first time it is
	 * encountered.  When the network traversal leaves a box on a
	 * connector that is higher than the connector on which it entered,
	 * the boxes it enters are assigned a number on the positive end of
	 * their row ordering; when the network traversal leaves a box on
	 * a connector that is lower than the connector on which it entered,
	 * the boxes it enters are assigned a number on the negative end of
	 * their row ordering.  For a simple tree-structured network, this
	 * algorithm gives a unique relative ordering of boxes within each
	 * row.  When the network is not a simple tree, the ordering can be
	 * strongly affected by the choice of the initial box.
	 * @param box the initial box for the calculation.
	 *        This only matters if the network is not a simple tree.
	 */
	public void rowOrder(Box box) {
		for (int b=0; b<boxes.length; b++)
			boxes[b].clearRowOrder();
		for (int n=0; n<nets.length; n++)
			nets[n].clearRowOrder();
		for (int r=0; r<rows.length; r++)
			rows[r].clearRowOrder();

		/* If there is a preferred root for placement, start with it. */
		if (box!=null)
			box.setRowOrder(false);
		/* Now make sure all other boxes are ordered. */
		for (int b=0; b<boxes.length; b++)
			boxes[b].setRowOrder(false);
	}

	/** Calculate the R position of each row.
	 */
	public void rowPosition() {
		/* Set the position of the boxes within each row (R coord)
		 * and calculate the length of each row. */
		for (int r=0; r<rows.length; r++)
			rows[r].normalize();

		/* Get the length of the longest row. */
		int maxLength = -1;
		int rowi = -1;		/* index of longest row */
		for (int r=0; r<rows.length; r++) {
			Row row = rows[r];
			if (!row.isBoxRow())
				continue;
			int rl = row.getLength();
			if (rl>maxLength) {
				maxLength = rl;
				rowi = r;
			}
		}
		maxRowLength = maxLength;

		/* For rows that are shorter than the longest row,
		 * we may be able to spread out the boxes to minimize
		 * line lengths.
		 */
		int r;
		for (r=rowi-1; r>=0; r--)
			rows[r].balanceAndSpread(NorE);
		for (r=rowi+1; r<rows.length; r++)
			rows[r].balanceAndSpread(SorW);
	}

	/** Do the channel route. */
	public void route() {
		/* Set the channel size for each row. */
		for (int r=0; r<rows.length; r++)
			rows[r].setChannelSize();
		
		/* Reposition the S coordinate of the rows based on the
		 * channel sizes. */
		int s = 0;
		for (int r=0; r<rows.length; r++) {
			rows[r].setS(s);
			s += rows[r].getThickness();
		}

		/* Update the S position on the boxes. */
		for (int b=0; b<boxes.length; b++)
			boxes[b].setS();
		
		/* Do the actual channel routing for each row. */
		for (int r=0; r<rows.length; r++)
			rows[r].route();
	}

	/** Calculate the bounding rectangle
	 * for all paintable items in the layout. */
	protected void setBounds() {
		BoundsRectangle r = new BoundsRectangle();
		if (boxes.length>0) {
			Box box = boxes[0];
			r.x = box.origin.x;
			r.y = box.origin.y;
			r.width = box.size.width;
			r.height = box.size.height;
		}
		for (int b=1; b<boxes.length; b++)
			boxes[b].addToBounds(r);
		for (int n=0; n<nets.length; n++)
			nets[n].addToBounds(r);
		bounds = r;
	}

	/** Get the bounds of the layout. */
	public BoundsRectangle getBounds() {
		return new BoundsRectangle(bounds);
	}

	/** Get the text line around the specified point. */
	public LayoutHitText getHitText(int x, int y) {
		/* Find the row with this point, get the text from it. */
		for (int r=0; r<rows.length; r++) {
			if (rows[r].contains(x,y)) {
				return rows[r].getHitText(x,y);
			}
		}
		return null;	/* not in any row */
	}

	/** Draw the layout. */
	public void paint(LayoutGraphics g) {
		/* Draw all the boxes. */
		for (int b=0; b<boxes.length; b++)
			boxes[b].paint(g);
		
		/* Draw all the connecting lines on the nets. */
		for (int n=0; n<nets.length; n++)
			nets[n].paint(g);
	}

	/** For debugging, dump the layout.
	 */
	public void dump(PrintWriter f) {
		f.println("Layout dump");
		f.println("  bounds="+bounds);
		dumpBoxes(f);
		dumpConnectors(f);
		dumpNets(f);
		dumpRows(f);
		f.flush();
	}

	/** Dump the boxes. */
	protected void dumpBoxes(PrintWriter f) {
		Box.dump(f,boxes," ",true);
	}

	/** Dump the connectors. */
	protected void dumpConnectors(PrintWriter f) {
		Connector.dump(f,connectors," ",true);
	}

	/** Dump the nets. */
	protected void dumpNets(PrintWriter f) {
		Net.dump(f,nets," ",true);
	}

	/** Dump the rows. */
	protected void dumpRows(PrintWriter f) {
		Row.dump(f,rows," ",true);
	}
}

/* end */
