/* LayoutGraphics.java
 *
 * Jim McBeath, July 12, 1998
 */

package jimmc.treepar;

import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;

/** LayoutGraphics holds a Graphics object and a transformation between
 * the Layout and the window in which it is being drawn.
 */
public class LayoutGraphics {
	/** Our Graphics object. */
	protected Graphics graphics;

	/** The bounds of the layout we are drawing. */
	protected Rectangle layoutBounds;

	/** The bounds of the window area we are drawing into. */
	protected Rectangle windowBounds;

	/** Numerator of scaling fraction for x. */
	protected int xScaleUp;

	/** Denominator of scaling fraction for x. */
	protected int xScaleDown;

	/** Post-scaling translation for x. */
	protected int xTranslate;

	/** Numerator of scaling fraction for y. */
	protected int yScaleUp;

	/** Denominator of scaling fraction for y. */
	protected int yScaleDown;

	/** Post-scaling translation for y. */
	protected int yTranslate;

	/** Create a new LayoutGraphics. */
	public LayoutGraphics(Graphics graphics) {
		this.graphics = graphics;
	}

	/** Set the size of the window into which we are drawing.
	 * @param sz the size of the window; we use the whole window
	 *        for drawing.
	 */
	public void setWindowSize(Dimension d) {
		setWindowBounds(new Rectangle(0,0,d.width,d.height));
	}

	/** Set the bounds of the window into which we are drawing.
	 * @param r the region of the window into which we should draw.
	 */
	public void setWindowBounds(Rectangle r) {
		windowBounds = r;
		calculateTransform();
	}

	/** Set the bounds of the Layout we are drawing.
	 * @param r the bounds of the layout.
	 */
	public void setLayoutBounds(Rectangle r) {
		layoutBounds = r;
		calculateTransform();
	}

	/** Calculate the transformation we use on points to take them from
	 * the Layout coordinate system to the window coordinate system.
	 */
	protected void calculateTransform() {
		if (layoutBounds==null || windowBounds==null)
			return;		/* not ready for this yet */

		/* Calculate the scaling factors for x. */
		xScaleUp = windowBounds.width;
		xScaleDown = layoutBounds.width;
		if (xScaleDown==0) {
			/* Can't transform */
			xScaleUp = 1;
			xScaleDown = 1;
		}

		/* Calculate the scaling factors for y. */
		yScaleUp = windowBounds.height;
		yScaleDown = layoutBounds.height;
		if (yScaleDown==0) {
			/* Can't transform */
			yScaleUp = 1;
			yScaleDown = 1;
		}

		/* Figure out which dimension has a larger scaling factor,
		 * shrink that one to match the smaller in order to preserve
		 * a one-to-one aspect ratio in the original coordinates.
		 */
		double xScale = ((double)xScaleUp)/(double)xScaleDown;
		double yScale = ((double)yScaleUp)/(double)yScaleDown;
		if (xScale>yScale) {
			/* need to decrease x scaling */
			xScaleDown = (int)((xScale/yScale)*(double)xScaleDown);
		} else if (yScale>xScale) {
			/* need to decrease y scaling */
			yScaleDown = (int)((yScale/xScale)*(double)yScaleDown);
		}

		/* Calcuate the post-scaling translation values for x and y. */
		xTranslate = windowBounds.x -
				layoutBounds.x*xScaleUp/xScaleDown;
		yTranslate = windowBounds.y -
				layoutBounds.y*yScaleUp/yScaleDown;
		
		/* Center in the window along the smaller-scaled dimension. */
		if (xScale>yScale) {
			xTranslate += (1.0-yScale/xScale)*0.5
					* windowBounds.width;
		} else {
			yTranslate += (1.0-xScale/yScale)*0.5
					* windowBounds.height;
		}

		/* The y-coordinates of the screen have the origin at the
		 * top, whereas the y-coordinates of the layout have the
		 * origin at the bottom.  We need to invert these as
		 * part of our Y transformation, so modify the scale and
		 * translate values to do so.
		 */
		yScaleUp = -yScaleUp;
		yTranslate = 2*windowBounds.y + windowBounds.height
				- yTranslate;
	}

	/** Set a font to render character the right size. */
	public void setFont(int lineHeight) {
		int screenSize = lineHeight * yScaleUp / yScaleDown;
		if (screenSize<0)
			screenSize = -screenSize;
		else if (screenSize==0)
			screenSize = 1;
		Font font = new Font("monospaced",Font.PLAIN,screenSize);
		graphics.setFont(font);
	}

	/** Transform an X value through our coordinate transform. */
	protected int transformX(int x) {
		return x * xScaleUp / xScaleDown + xTranslate;
	}

	/** Transform an Y value through our coordinate transform. */
	protected int transformY(int y) {
		return y * yScaleUp / yScaleDown + yTranslate;
	}

	/** Reverse-transform an X value to convert screen coordinates
	 * to layout coordinates. */
	protected int reverseTransformX(int x) {
		return (x - xTranslate)*xScaleDown/xScaleUp;
	}

	/** Reverse-transform a Y value to convert screen coordinates
	 * to layout coordinates. */
	protected int reverseTransformY(int y) {
		return (y - yTranslate)*yScaleDown/yScaleUp;
	}

	/** Draw a line between two points. */
	public void drawLine(Point p0, Point p1) {
		drawLine(p0.x, p0.y, p1.x, p1.y);
	}

	/** Draw a line. */
	public void drawLine(int x0, int y0, int x1, int y1) {
		graphics.drawLine(transformX(x0),transformY(y0),
				  transformX(x1),transformY(y1));
	}

	/** Draw the outline of a box. */
	public void drawBox(Point origin, Dimension size) {
		int lx = origin.x;
		int ly = origin.y;
		int hx = origin.x + size.width;
		int hy = origin.y + size.height;
		drawLine(lx,ly,lx,hy);	/* lower left to upper left */
		drawLine(lx,hy,hx,hy);	/* upper left to upper right */
		drawLine(hx,hy,hx,ly);	/* upper right to lower right */
		drawLine(hx,ly,lx,ly);	/* lower right to lower left */
	}

	/** Draw some text. */
	public void drawString(String text, Point p) {
		drawString(text, p.x, p.y);
	}

	/** Draw some text. */
	public void drawString(String text, int x, int y) {
		if (text==null)
			return;		/* ignore null strings */
		graphics.drawString(text,transformX(x),transformY(y));
	}
}

/* end */
